@@ -1,7 +1,7 @@
 #!/usr/bin/env python3
 """Api client code for the Adnuntius APIs."""
 
-__copyright__ = "Copyright (c) 2021 Adnuntius AS.  All rights reserved."
+__copyright__ = "Copyright (c) 2022 Adnuntius AS.  All rights reserved."
 
 import json
 import os
@@ -128,6 +128,7 @@ class Api:
         self.location_stats = api_client("stats/location/impression")
         self.targeting_templates = api_client("targetingtemplates")
         self.teams = api_client("teams")
+        self.teamgroups = api_client("teamgroups")
         self.tiers = api_client("tiers")
         self.timezones = api_client("timezones")
         self.traffic = api_client("stats/traffic")
@@ -156,6 +157,7 @@ class ApiClient:
         """
         self.resourceName = resource_name
         self.api = api_context
+        self.two_factor_code_provider = None
         self.authorisation = None
         self.auth_time = None
         self.refresh_token = None
@@ -358,6 +360,40 @@ class ApiClient:
         else:
             return r.text
 
+    def __do_two_factor_auth(self):
+        if hasattr(self.api, 'two_factor_code_provider'):
+            code = self.api.two_factor_code_provider()
+        else:
+            self.authorisation = None
+            raise RuntimeError("2FA setup failure: api.two_factor_code_provider not defined")
+
+        data = {'code': code}
+        endpoint = "/authenticate/2fa"
+
+        headers = {'Content-Type': 'application/json'}
+        headers.update(self.api.headers)
+        headers.update(self.authorisation)
+
+        try:
+            r = self.handle_err(self.session.post(self.baseUrl + endpoint, data=json.dumps(data),
+                                params=self.api.defaultAuthArgs, headers=headers))
+        except RuntimeError as e:
+            # for a failed 2fa, we need to clear the 2FA authorisation field
+            # the api will retry the entire auth process again for the next call
+            self.authorisation = None
+            raise e
+
+        response = r.json()
+        # a normal Authentication failure will already have been raised above, this catches
+        # a unexpected situation where there is no access token
+        if 'access_token' not in response:
+            self.authorisation = None
+            raise RuntimeError("Unexpected 2FA authentication failure in POST " + r.url)
+
+        self.authorisation = {'Authorization': 'Bearer ' + response['access_token']}
+        self.auth_time = time.time()
+        self.refresh_token = response['refresh_token']
+
     def __do_password_auth(self):
         data = {'grant_type': 'password',
                 'scope': 'ng_api',
@@ -375,12 +411,21 @@ class ApiClient:
 
         r = self.handle_err(self.session.post(self.baseUrl + endpoint, data=json.dumps(data),
                                               params=self.api.defaultAuthArgs, headers=headers))
+
         response = r.json()
+        # a normal Authentication failure will already have been raised above, this catches
+        # a unexpected situation where there is no access token
         if 'access_token' not in response:
-            raise RuntimeError("API authentication failed in POST " + r.url)
+            self.authorisation = None
+            raise RuntimeError("Unexpected API authentication failed in POST " + r.url)
+
         self.authorisation = {'Authorization': 'Bearer ' + response['access_token']}
-        self.auth_time = time.time()
-        self.refresh_token = response['refresh_token']
+
+        if 'scope' in response and response['scope'] == 'TWO_FACTOR_AUTH':
+            self.__do_two_factor_auth()
+        else:
+            self.auth_time = time.time()
+            self.refresh_token = response['refresh_token']
 
     def __do_refresh_token_auth(self):
         data = {'grant_type': 'refresh_token',
@@ -411,7 +456,7 @@ class ApiClient:
         """
 
         # if we have an existing authorisation but its approaching one hour of age, discard it and refresh.
-        if self.authorisation and not self.api.apiKey and self.api.username:
+        if self.authorisation and not self.api.apiKey and self.api.username and self.auth_time:
             current_time = time.time()
 
             if current_time - self.auth_time > AUTH_TOKEN_SAFE_EXPIRY_IN_SECS:
@@ -729,15 +774,16 @@ class AdServer:
         r = self.session.post(self.__get_base_url(headers) + "/pixelc.gif", data=json.dumps(data), headers=headers)
         return r
 
-    def trigger_event(self, url):
+    def trigger_event(self, url, headers=None):
         """
         Triggers an event by requesting a URL. Uses the ad-server session so that cookies are shared.
-        :param url:
+        :param url: the url to request
+        :param headers: any additional headers to add to the request
         :return:
         """
         if url[0:2] == '//':
             url = "http:" + url
-        r = self.session.get(url, allow_redirects=False)
+        r = self.session.get(url, allow_redirects=False, headers=headers)
         return r
 
     def post_event(self, url, event):
